@@ -6,37 +6,51 @@ This module contains code for the global BlueZ D-Bus object manager that is
 used internally by Bleak.
 """
 
+import sys
+from typing import TYPE_CHECKING
+
+if TYPE_CHECKING:
+    if sys.platform != "linux":
+        assert False, "This backend is only available on Linux"
+
 import asyncio
+import contextlib
 import logging
 import os
-from typing import (
-    Any,
-    Callable,
-    Coroutine,
-    Dict,
-    Iterable,
-    List,
-    MutableMapping,
-    NamedTuple,
-    Optional,
-    Set,
-    cast,
-)
+from collections import defaultdict
+from collections.abc import Callable, Coroutine, MutableMapping
+from functools import partial
+from typing import Any, NamedTuple, Optional, cast
 from weakref import WeakKeyDictionary
 
-from dbus_fast import BusType, Message, MessageType, Variant, unpack_variants
+from dbus_fast import AuthError, BusType, Message, MessageType, Variant, unpack_variants
 from dbus_fast.aio.message_bus import MessageBus
 
-from ...exc import BleakError
-from ..service import BleakGATTServiceCollection
-from . import defs
-from .advertisement_monitor import AdvertisementMonitor, OrPatternLike
-from .characteristic import BleakGATTCharacteristicBlueZDBus
-from .defs import Device1, GattService1, GattCharacteristic1, GattDescriptor1
-from .descriptor import BleakGATTDescriptorBlueZDBus
-from .service import BleakGATTServiceBlueZDBus
-from .signals import MatchRules, add_match
-from .utils import assert_reply
+from bleak.args.bluez import OrPatternLike
+from bleak.backends.bluezdbus import defs
+from bleak.backends.bluezdbus.advertisement_monitor import AdvertisementMonitor
+from bleak.backends.bluezdbus.defs import (
+    Device1,
+    GattCharacteristic1,
+    GattDescriptor1,
+    GattService1,
+)
+from bleak.backends.bluezdbus.signals import MatchRules, add_match
+from bleak.backends.bluezdbus.utils import (
+    assert_reply,
+    device_path_from_characteristic_path,
+    extract_service_handle_from_path,
+    get_dbus_authenticator,
+)
+from bleak.backends.characteristic import BleakGATTCharacteristic
+from bleak.backends.descriptor import BleakGATTDescriptor
+from bleak.backends.service import BleakGATTService, BleakGATTServiceCollection
+from bleak.exc import (
+    BleakBluetoothNotAvailableError,
+    BleakBluetoothNotAvailableReason,
+    BleakDBusError,
+    BleakError,
+)
 
 logger = logging.getLogger(__name__)
 
@@ -50,19 +64,28 @@ Args:
 """
 
 
-class CallbackAndState(NamedTuple):
+DevicePropertiesChangedCallback = Callable[[Optional[Any]], None]
+"""
+A callback that is called when the properties of a device change in BlueZ.
+
+Args:
+    arg0: The new property value.
+"""
+
+
+class DeviceConditionCallback(NamedTuple):
     """
-    Encapsulates an :data:`AdvertisementCallback` and some state.
+    Encapsulates a :data:`DevicePropertiesChangedCallback` and the property name being watched.
     """
 
-    callback: AdvertisementCallback
+    callback: DevicePropertiesChangedCallback
     """
     The callback.
     """
 
-    adapter_path: str
+    property_name: str
     """
-    The D-Bus object path of the adapter associated with the callback.
+    The name of the property to watch.
     """
 
 
@@ -110,7 +133,6 @@ Args:
 
 
 class DeviceWatcher(NamedTuple):
-
     device_path: str
     """
     The D-Bus object path of the device.
@@ -138,6 +160,12 @@ _ADVERTISING_DATA_PROPERTIES = {
 }
 
 
+def get_max_write_without_response_size(char_props: GattCharacteristic1) -> int:
+    # "MTU" property was added in BlueZ 5.62, otherwise fall
+    # back to minimum MTU according to Bluetooth spec.
+    return char_props.get("MTU", 23) - 3
+
+
 class BlueZManager:
     """
     BlueZ D-Bus object manager.
@@ -145,33 +173,73 @@ class BlueZManager:
     Use :func:`bleak.backends.bluezdbus.get_global_bluez_manager` to get the global instance.
     """
 
-    def __init__(self):
+    def __init__(self) -> None:
         self._bus: Optional[MessageBus] = None
         self._bus_lock = asyncio.Lock()
 
         # dict of object path: dict of interface name: dict of property name: property value
-        self._properties: Dict[str, Dict[str, Dict[str, Any]]] = {}
+        self._properties: dict[str, dict[str, dict[str, Any]]] = {}
 
         # set of available adapters for quick lookup
-        self._adapters: Set[str] = set()
+        self._adapters: set[str] = set()
 
         # The BlueZ APIs only maps children to parents, so we need to keep maps
         # to quickly find the children of a parent D-Bus object.
 
         # map of device d-bus object paths to set of service d-bus object paths
-        self._service_map: Dict[str, Set[str]] = {}
+        self._service_map: dict[str, set[str]] = {}
         # map of service d-bus object paths to set of characteristic d-bus object paths
-        self._characteristic_map: Dict[str, Set[str]] = {}
+        self._characteristic_map: dict[str, set[str]] = {}
         # map of characteristic d-bus object paths to set of descriptor d-bus object paths
-        self._descriptor_map: Dict[str, Set[str]] = {}
+        self._descriptor_map: dict[str, set[str]] = {}
 
-        self._advertisement_callbacks: List[CallbackAndState] = []
-        self._device_removed_callbacks: List[DeviceRemovedCallbackAndState] = []
-        self._device_watchers: Set[DeviceWatcher] = set()
-        self._condition_callbacks: Set[Callable] = set()
-        self._services_cache: Dict[str, BleakGATTServiceCollection] = {}
+        self._advertisement_callbacks: defaultdict[str, list[AdvertisementCallback]] = (
+            defaultdict(list)
+        )
+        self._device_removed_callbacks: list[DeviceRemovedCallbackAndState] = []
+        self._device_watchers: dict[str, set[DeviceWatcher]] = {}
+        self._condition_callbacks: dict[str, set[DeviceConditionCallback]] = {}
+        self._services_cache: dict[str, BleakGATTServiceCollection] = {}
 
-    async def async_init(self):
+    def _check_adapter(self, adapter_path: str) -> None:
+        """
+        Raises:
+            BleakError: if adapter is not present in BlueZ
+        """
+        if adapter_path not in self._properties:
+            raise BleakError(f"adapter '{adapter_path.split('/')[-1]}' not found")
+
+    def _check_device(self, device_path: str) -> None:
+        """
+        Raises:
+            BleakError: if device is not present in BlueZ
+        """
+        if device_path not in self._properties:
+            raise BleakError(f"device '{device_path.split('/')[-1]}' not found")
+
+    def _get_device_property(
+        self, device_path: str, interface: str, property_name: str
+    ) -> Any:
+        self._check_device(device_path)
+        device_properties = self._properties[device_path]
+
+        try:
+            interface_properties = device_properties[interface]
+        except KeyError:
+            raise BleakError(
+                f"Interface {interface} not found for device '{device_path}'"
+            )
+
+        try:
+            value = interface_properties[property_name]
+        except KeyError:
+            raise BleakError(
+                f"Property '{property_name}' not found for '{interface}' in '{device_path}'"
+            )
+
+        return value
+
+    async def async_init(self) -> None:
         """
         Connects to the D-Bus message bus and begins monitoring signals.
 
@@ -185,15 +253,25 @@ class BlueZManager:
             self._services_cache = {}
 
             # We need to create a new MessageBus each time as
-            # dbus-next will destory the underlying file descriptors
+            # dbus-next will destroy the underlying file descriptors
             # when the previous one is closed in its finalizer.
-            bus = MessageBus(bus_type=BusType.SYSTEM)
-            await bus.connect()
+            bus = MessageBus(bus_type=BusType.SYSTEM, auth=get_dbus_authenticator())
 
             try:
+                # We need to call bus.disconnect() even when bus.connect() fails in
+                # order to release the file handles created in the constructor.
+                try:
+                    await bus.connect()
+                except AuthError as e:
+                    raise BleakBluetoothNotAvailableError(
+                        e.args[0],
+                        BleakBluetoothNotAvailableReason.DENIED_BY_SYSTEM,
+                    ) from e
+
                 # Add signal listeners
 
                 bus.add_message_handler(self._parse_msg)
+                reply: Optional[Message]
 
                 rules = MatchRules(
                     interface=defs.OBJECT_MANAGER_INTERFACE,
@@ -274,13 +352,19 @@ class BlueZManager:
                             desc_props["Characteristic"], set()
                         ).add(path)
 
-                logger.debug(f"initial properties: {self._properties}")
+                if logger.isEnabledFor(logging.DEBUG):
+                    logger.debug("initial properties: %s", self._properties)
 
             except BaseException:
                 # if setup failed, disconnect
                 bus.disconnect()
                 raise
 
+            if self._bus:
+                # Even if we are disconnected, still need to call this to
+                # release file handles.
+                self._bus.disconnect()
+
             # Everything is setup, so save the bus
             self._bus = bus
 
@@ -292,27 +376,50 @@ class BlueZManager:
             Name of the first found powered adapter on the system, i.e. "/org/bluez/hciX".
 
         Raises:
-            BleakError:
-                if there are no Bluetooth adapters or if none of the adapters are powered
+            BleakBluetoothNotAvailableError:
+                if there are no Bluetooth Low Energy adapters or if none of the adapters are powered
+
+        .. versionchanged:: 2.0
+            Now raises :class:`BleakBluetoothNotAvailableError` instead of :class:`BleakError`.
         """
         if not any(self._adapters):
-            raise BleakError("No Bluetooth adapters found.")
+            raise BleakBluetoothNotAvailableError(
+                "No Bluetooth adapters found.",
+                BleakBluetoothNotAvailableReason.NO_BLUETOOTH,
+            )
 
-        for adapter_path in self._adapters:
+        ble_central_adapters = list(
+            filter(
+                lambda a: "central"
+                in self._properties[a][defs.ADAPTER_INTERFACE]["Roles"],
+                self._adapters,
+            )
+        )
+
+        if not ble_central_adapters:
+            raise BleakBluetoothNotAvailableError(
+                "No Bluetooth adapters with BLE 'central' role found.",
+                BleakBluetoothNotAvailableReason.NO_BLE_CENTRAL_ROLE,
+            )
+
+        for adapter_path in ble_central_adapters:
             if cast(
                 defs.Adapter1, self._properties[adapter_path][defs.ADAPTER_INTERFACE]
             )["Powered"]:
                 return adapter_path
 
-        raise BleakError("No powered Bluetooth adapters found.")
+        raise BleakBluetoothNotAvailableError(
+            "No powered Bluetooth adapters found. Turn on Bluetooth and try again.",
+            BleakBluetoothNotAvailableReason.POWERED_OFF,
+        )
 
     async def active_scan(
         self,
         adapter_path: str,
-        filters: Dict[str, Variant],
+        filters: dict[str, Variant],
         advertisement_callback: AdvertisementCallback,
         device_removed_callback: DeviceRemovedCallback,
-    ) -> Callable[[], Coroutine]:
+    ) -> Callable[[], Coroutine[Any, Any, None]]:
         """
         Configures the advertisement data filters and starts scanning.
 
@@ -326,16 +433,19 @@ class BlueZManager:
 
         Returns:
             An async function that is used to stop scanning and remove the filters.
+
+        Raises:
+            BleakError: if the adapter is not present in BlueZ
         """
         async with self._bus_lock:
+            assert self._bus
+
             # If the adapter doesn't exist, then the message calls below would
             # fail with "method not found". This provides a more informative
             # error message.
-            if adapter_path not in self._properties:
-                raise BleakError(f"adapter '{adapter_path.split('/')[-1]}' not found")
+            self._check_adapter(adapter_path)
 
-            callback_and_state = CallbackAndState(advertisement_callback, adapter_path)
-            self._advertisement_callbacks.append(callback_and_state)
+            self._advertisement_callbacks[adapter_path].append(advertisement_callback)
 
             device_removed_callback_and_state = DeviceRemovedCallbackAndState(
                 device_removed_callback, adapter_path
@@ -371,12 +481,16 @@ class BlueZManager:
                     # need to remove callbacks first, otherwise we get TxPower
                     # and RSSI properties removed during stop which causes
                     # incorrect advertisement data callbacks
-                    self._advertisement_callbacks.remove(callback_and_state)
+                    self._advertisement_callbacks[adapter_path].remove(
+                        advertisement_callback
+                    )
                     self._device_removed_callbacks.remove(
                         device_removed_callback_and_state
                     )
 
                     async with self._bus_lock:
+                        assert self._bus
+
                         reply = await self._bus.call(
                             Message(
                                 destination=defs.BLUEZ_SERVICE,
@@ -385,35 +499,42 @@ class BlueZManager:
                                 member="StopDiscovery",
                             )
                         )
-                        assert_reply(reply)
 
-                        # remove the filters
-                        reply = await self._bus.call(
-                            Message(
-                                destination=defs.BLUEZ_SERVICE,
-                                path=adapter_path,
-                                interface=defs.ADAPTER_INTERFACE,
-                                member="SetDiscoveryFilter",
-                                signature="a{sv}",
-                                body=[{}],
+                        try:
+                            assert_reply(reply)
+                        except BleakDBusError as ex:
+                            if ex.dbus_error != "org.bluez.Error.NotReady":
+                                raise
+                        else:
+                            # remove the filters
+                            reply = await self._bus.call(
+                                Message(
+                                    destination=defs.BLUEZ_SERVICE,
+                                    path=adapter_path,
+                                    interface=defs.ADAPTER_INTERFACE,
+                                    member="SetDiscoveryFilter",
+                                    signature="a{sv}",
+                                    body=[{}],
+                                )
                             )
-                        )
-                        assert_reply(reply)
+                            assert_reply(reply)
 
                 return stop
             except BaseException:
                 # if starting scanning failed, don't leak the callbacks
-                self._advertisement_callbacks.remove(callback_and_state)
+                self._advertisement_callbacks[adapter_path].remove(
+                    advertisement_callback
+                )
                 self._device_removed_callbacks.remove(device_removed_callback_and_state)
                 raise
 
     async def passive_scan(
         self,
         adapter_path: str,
-        filters: List[OrPatternLike],
+        filters: list[OrPatternLike],
         advertisement_callback: AdvertisementCallback,
         device_removed_callback: DeviceRemovedCallback,
-    ) -> Callable[[], Coroutine]:
+    ) -> Callable[[], Coroutine[Any, Any, None]]:
         """
         Configures the advertisement data filters and starts scanning.
 
@@ -427,16 +548,19 @@ class BlueZManager:
 
         Returns:
             An async function that is used to stop scanning and remove the filters.
+
+        Raises:
+            BleakError: if the adapter is not present in BlueZ
         """
         async with self._bus_lock:
+            assert self._bus
+
             # If the adapter doesn't exist, then the message calls below would
             # fail with "method not found". This provides a more informative
             # error message.
-            if adapter_path not in self._properties:
-                raise BleakError(f"adapter '{adapter_path.split('/')[-1]}' not found")
+            self._check_adapter(adapter_path)
 
-            callback_and_state = CallbackAndState(advertisement_callback, adapter_path)
-            self._advertisement_callbacks.append(callback_and_state)
+            self._advertisement_callbacks[adapter_path].append(advertisement_callback)
 
             device_removed_callback_and_state = DeviceRemovedCallbackAndState(
                 device_removed_callback, adapter_path
@@ -466,7 +590,7 @@ class BlueZManager:
                     and reply.error_name == "org.freedesktop.DBus.Error.UnknownMethod"
                 ):
                     raise BleakError(
-                        "passive scanning on Linux requires BlueZ >= 5.55 with --experimental enabled and Linux kernel >= 5.10"
+                        "passive scanning on Linux requires BlueZ >= 5.56 with --experimental enabled and Linux kernel >= 5.10"
                     )
 
                 assert_reply(reply)
@@ -475,16 +599,20 @@ class BlueZManager:
                 # won't use the monitor
                 self._bus.export(monitor_path, monitor)
 
-                async def stop():
+                async def stop() -> None:
                     # need to remove callbacks first, otherwise we get TxPower
                     # and RSSI properties removed during stop which causes
                     # incorrect advertisement data callbacks
-                    self._advertisement_callbacks.remove(callback_and_state)
+                    self._advertisement_callbacks[adapter_path].remove(
+                        advertisement_callback
+                    )
                     self._device_removed_callbacks.remove(
                         device_removed_callback_and_state
                     )
 
                     async with self._bus_lock:
+                        assert self._bus
+
                         self._bus.unexport(monitor_path, monitor)
 
                         reply = await self._bus.call(
@@ -503,7 +631,9 @@ class BlueZManager:
 
             except BaseException:
                 # if starting scanning failed, don't leak the callbacks
-                self._advertisement_callbacks.remove(callback_and_state)
+                self._advertisement_callbacks[adapter_path].remove(
+                    advertisement_callback
+                )
                 self._device_removed_callbacks.remove(device_removed_callback_and_state)
                 raise
 
@@ -529,12 +659,17 @@ class BlueZManager:
 
         Returns:
             A device watcher object that acts a token to unregister the watcher.
+
+        Raises:
+            BleakError: if the device is not present in BlueZ
         """
+        self._check_device(device_path)
+
         watcher = DeviceWatcher(
             device_path, on_connected_changed, on_characteristic_value_changed
         )
 
-        self._device_watchers.add(watcher)
+        self._device_watchers.setdefault(device_path, set()).add(watcher)
         return watcher
 
     def remove_device_watcher(self, watcher: DeviceWatcher) -> None:
@@ -545,10 +680,13 @@ class BlueZManager:
             The device watcher token that was returned by
             :meth:`add_device_watcher`.
         """
-        self._device_watchers.remove(watcher)
+        device_path = watcher.device_path
+        self._device_watchers[device_path].remove(watcher)
+        if not self._device_watchers[device_path]:
+            del self._device_watchers[device_path]
 
     async def get_services(
-        self, device_path: str, use_cached: bool
+        self, device_path: str, use_cached: bool, requested_services: Optional[set[str]]
     ) -> BleakGATTServiceCollection:
         """
         Builds a new :class:`BleakGATTServiceCollection` from the current state.
@@ -560,10 +698,18 @@ class BlueZManager:
                 When ``True`` if there is a cached :class:`BleakGATTServiceCollection`,
                 the method will not wait for ``"ServicesResolved"`` to become true
                 and instead return the cached service collection immediately.
+            requested_services:
+                When given, only return services with UUID that is in the list
+                of requested services.
 
         Returns:
             A new :class:`BleakGATTServiceCollection`.
+
+        Raises:
+            BleakError: if the device is not present in BlueZ
         """
+        self._check_device(device_path)
+
         if use_cached:
             services = self._services_cache.get(device_path)
             if services is not None:
@@ -580,7 +726,17 @@ class BlueZManager:
                 self._properties[service_path][defs.GATT_SERVICE_INTERFACE],
             )
 
-            service = BleakGATTServiceBlueZDBus(service_props, service_path)
+            service = BleakGATTService(
+                (service_path, service_props),
+                extract_service_handle_from_path(service_path),
+                service_props["UUID"],
+            )
+
+            if (
+                requested_services is not None
+                and service.uuid not in requested_services
+            ):
+                continue
 
             services.add_service(service)
 
@@ -590,14 +746,17 @@ class BlueZManager:
                     self._properties[char_path][defs.GATT_CHARACTERISTIC_INTERFACE],
                 )
 
-                char = BleakGATTCharacteristicBlueZDBus(
-                    char_props,
-                    char_path,
-                    service.uuid,
-                    service.handle,
-                    # "MTU" property was added in BlueZ 5.62, otherwise fall
-                    # back to minimum MTU according to Bluetooth spec.
-                    char_props.get("MTU", 23) - 3,
+                char = BleakGATTCharacteristic(
+                    (char_path, char_props),
+                    extract_service_handle_from_path(char_path),
+                    char_props["UUID"],
+                    char_props["Flags"],
+                    # Because `char_props` is a loop varialbe, we cannot
+                    # directly bind a closure (i.e. lambda) to it;
+                    # instead, we let `functools.partial` create a new
+                    # function frame to close over at each iteration.
+                    partial(get_max_write_without_response_size, char_props),
+                    service,
                 )
 
                 services.add_characteristic(char)
@@ -608,11 +767,11 @@ class BlueZManager:
                         self._properties[desc_path][defs.GATT_DESCRIPTOR_INTERFACE],
                     )
 
-                    desc = BleakGATTDescriptorBlueZDBus(
-                        desc_props,
-                        desc_path,
-                        char.uuid,
-                        char.handle,
+                    desc = BleakGATTDescriptor(
+                        (desc_path, desc_props),
+                        int(desc_path[-4:], 16),
+                        desc_props["UUID"],
+                        char,
                     )
 
                     services.add_descriptor(desc)
@@ -630,8 +789,26 @@ class BlueZManager:
 
         Returns:
             The current property value.
+
+        Raises:
+            BleakError: if the device is not present in BlueZ
         """
-        return self._properties[device_path][defs.DEVICE_INTERFACE]["Name"]
+        return self._get_device_property(device_path, defs.DEVICE_INTERFACE, "Name")
+
+    def get_device_address(self, device_path: str) -> str:
+        """
+        Gets the value of the "Address" property for a device.
+
+        Args:
+            device_path: The D-Bus object path of the device.
+
+        Returns:
+            The current property value.
+
+        Raises:
+            BleakError: if the device is not present in BlueZ
+        """
+        return self._get_device_property(device_path, defs.DEVICE_INTERFACE, "Address")
 
     def is_connected(self, device_path: str) -> bool:
         """
@@ -641,35 +818,104 @@ class BlueZManager:
             device_path: The D-Bus object path of the device.
 
         Returns:
-            The current property value.
+            The current property value or ``False`` if the device does not exist in BlueZ.
         """
         try:
             return self._properties[device_path][defs.DEVICE_INTERFACE]["Connected"]
         except KeyError:
             return False
 
+    def is_paired(self, device_path: str) -> bool:
+        """
+        Gets the value of the "Paired" property for a device.
+
+        Args:
+            device_path: The D-Bus object path of the device.
+
+        Returns:
+            The current property value or ``False`` if the device does not exist in BlueZ.
+        """
+        try:
+            return self._properties[device_path][defs.DEVICE_INTERFACE]["Paired"]
+        except KeyError:
+            return False
+
     async def _wait_for_services_discovery(self, device_path: str) -> None:
         """
         Waits for the device services to be discovered.
 
         If a disconnect happens before the completion a BleakError exception is raised.
+
+        Raises:
+            BleakError: if the device is not present in BlueZ
         """
-        services_discovered_wait_task = asyncio.create_task(
-            self._wait_condition(device_path, "ServicesResolved", True)
-        )
-        device_disconnected_wait_task = asyncio.create_task(
-            self._wait_condition(device_path, "Connected", False)
-        )
-        done, pending = await asyncio.wait(
-            {services_discovered_wait_task, device_disconnected_wait_task},
-            return_when=asyncio.FIRST_COMPLETED,
+        self._check_device(device_path)
+
+        with contextlib.ExitStack() as stack:
+            services_discovered_wait_task = asyncio.create_task(
+                self._wait_condition(device_path, "ServicesResolved", True)
+            )
+            stack.callback(services_discovered_wait_task.cancel)
+
+            device_disconnected_wait_task = asyncio.create_task(
+                self._wait_condition(device_path, "Connected", False)
+            )
+            stack.callback(device_disconnected_wait_task.cancel)
+
+            # in some cases, we can get "InterfaceRemoved" without the
+            # "Connected" property changing, so we need to race against both
+            # conditions
+            device_removed_wait_task = asyncio.create_task(
+                self._wait_removed(device_path)
+            )
+            stack.callback(device_removed_wait_task.cancel)
+
+            done, _ = await asyncio.wait(
+                {
+                    services_discovered_wait_task,
+                    device_disconnected_wait_task,
+                    device_removed_wait_task,
+                },
+                return_when=asyncio.FIRST_COMPLETED,
+            )
+
+            # check for exceptions
+            for task in done:
+                task.result()
+
+            if not done.isdisjoint(
+                {device_disconnected_wait_task, device_removed_wait_task}
+            ):
+                raise BleakError("failed to discover services, device disconnected")
+
+    async def _wait_removed(self, device_path: str) -> None:
+        """
+        Waits for the device interface to be removed.
+
+        If the device is not present in BlueZ, this returns immediately.
+
+        Args:
+            device_path: The D-Bus object path of a Bluetooth device.
+        """
+        if device_path not in self._properties:
+            return
+
+        event = asyncio.Event()
+
+        def callback(o: str) -> None:
+            if o == device_path:
+                event.set()
+
+        device_removed_callback_and_state = DeviceRemovedCallbackAndState(
+            callback, self._properties[device_path][defs.DEVICE_INTERFACE]["Adapter"]
         )
 
-        for p in pending:
-            p.cancel()
-
-        if device_disconnected_wait_task in done:
-            raise BleakError("failed to discover services, device disconnected")
+        with contextlib.ExitStack() as stack:
+            self._device_removed_callbacks.append(device_removed_callback_and_state)
+            stack.callback(
+                self._device_removed_callbacks.remove, device_removed_callback_and_state
+            )
+            await event.wait()
 
     async def _wait_condition(
         self, device_path: str, property_name: str, property_value: Any
@@ -681,31 +927,38 @@ class BlueZManager:
             device_path: The D-Bus object path of a Bluetooth device.
             property_name: The name of the property to test.
             property_value: A value to compare the current property value to.
+
+        Raises:
+            BleakError: if the device is not present in BlueZ
         """
-        if (
-            self._properties[device_path][defs.DEVICE_INTERFACE][property_name]
-            == property_value
-        ):
+        value = self._get_device_property(
+            device_path, defs.DEVICE_INTERFACE, property_name
+        )
+
+        if value == property_value:
             return
 
         event = asyncio.Event()
 
-        def callback():
-            if (
-                self._properties[device_path][defs.DEVICE_INTERFACE][property_name]
-                == property_value
-            ):
+        def _wait_condition_callback(new_value: Optional[Any]) -> None:
+            """Callback for when a property changes."""
+            if new_value == property_value:
                 event.set()
 
-        self._condition_callbacks.add(callback)
+        condition_callbacks = self._condition_callbacks
+        device_callbacks = condition_callbacks.setdefault(device_path, set())
+        callback = DeviceConditionCallback(_wait_condition_callback, property_name)
+        device_callbacks.add(callback)
 
         try:
             # can be canceled
             await event.wait()
         finally:
-            self._condition_callbacks.remove(callback)
+            device_callbacks.remove(callback)
+            if not device_callbacks:
+                del condition_callbacks[device_path]
 
-    def _parse_msg(self, message: Message):
+    def _parse_msg(self, message: Message) -> None:
         """
         Handles callbacks from dbus_fast.
         """
@@ -713,21 +966,22 @@ class BlueZManager:
         if message.message_type != MessageType.SIGNAL:
             return
 
-        logger.debug(
-            "received D-Bus signal: %s.%s (%s): %s",
-            message.interface,
-            message.member,
-            message.path,
-            message.body,
-        )
+        if logger.isEnabledFor(logging.DEBUG):
+            logger.debug(
+                "received D-Bus signal: %s.%s (%s): %s",
+                message.interface,
+                message.member,
+                message.path,
+                message.body,
+            )
 
         # type hints
         obj_path: str
-        interfaces_and_props: Dict[str, Dict[str, Variant]]
-        interfaces: List[str]
+        interfaces_and_props: dict[str, dict[str, Variant]]
+        interfaces: list[str]
         interface: str
-        changed: Dict[str, Variant]
-        invalidated: List[str]
+        changed: dict[str, Variant]
+        invalidated: list[str]
 
         if message.member == "InterfacesAdded":
             obj_path, interfaces_and_props = message.body
@@ -762,7 +1016,7 @@ class BlueZManager:
                 # devices that only advertise once and then go to sleep for a while.
                 elif interface == defs.DEVICE_INTERFACE:
                     self._run_advertisement_callbacks(
-                        obj_path, cast(Device1, unpacked_props), unpacked_props.keys()
+                        obj_path, cast(Device1, unpacked_props)
                     )
         elif message.member == "InterfacesRemoved":
             obj_path, interfaces = message.body
@@ -789,6 +1043,13 @@ class BlueZManager:
                         if obj_path.startswith(adapter_path):
                             callback(obj_path)
                 elif interface == defs.GATT_SERVICE_INTERFACE:
+                    device_path = obj_path[: obj_path.rfind("/")]
+
+                    try:
+                        self._service_map[device_path].remove(obj_path)
+                    except KeyError:
+                        pass
+
                     try:
                         del self._characteristic_map[obj_path]
                     except KeyError:
@@ -798,13 +1059,19 @@ class BlueZManager:
                         del self._descriptor_map[obj_path]
                     except KeyError:
                         pass
-        elif message.member == "PropertiesChanged":
-            assert message.path is not None
 
+            # Remove empty properties when all interfaces have been removed.
+            # This avoids wasting memory for people who have noisy devices
+            # with private addresses that change frequently.
+            if obj_path in self._properties and not self._properties[obj_path]:
+                del self._properties[obj_path]
+        elif message.member == "PropertiesChanged":
             interface, changed, invalidated = message.body
+            message_path = message.path
+            assert message_path is not None
 
             try:
-                self_interface = self._properties[message.path][interface]
+                self_interface = self._properties[message_path][interface]
             except KeyError:
                 # This can happen during initialization. The "PropertiesChanged"
                 # handler is attached before "GetManagedObjects" is called
@@ -818,58 +1085,63 @@ class BlueZManager:
                 self_interface.update(unpack_variants(changed))
 
                 for name in invalidated:
-                    del self_interface[name]
+                    try:
+                        del self_interface[name]
+                    except KeyError:
+                        # sometimes there BlueZ tries to remove properties
+                        # that were never added
+                        pass
 
                 # then call any callbacks so they will be called with the
                 # updated state
 
                 if interface == defs.DEVICE_INTERFACE:
                     # handle advertisement watchers
+                    device_path = message_path
 
                     self._run_advertisement_callbacks(
-                        message.path, cast(Device1, self_interface), changed.keys()
+                        device_path, cast(Device1, self_interface)
                     )
 
                     # handle device condition watchers
-                    for condition_callback in self._condition_callbacks:
-                        condition_callback()
+                    callbacks = self._condition_callbacks.get(device_path)
+                    if callbacks:
+                        for item in callbacks:
+                            name = item.property_name
+                            if name in changed:
+                                item.callback(self_interface.get(name))
 
                     # handle device connection change watchers
-
                     if "Connected" in changed:
-                        for (
-                            device_path,
-                            on_connected_changed,
-                            _,
-                        ) in self._device_watchers.copy():
-                            # callbacks may remove the watcher, hence the copy() above
-                            if message.path == device_path:
-                                on_connected_changed(self_interface["Connected"])
+                        new_connected = self_interface["Connected"]
+                        watchers = self._device_watchers.get(device_path)
+                        if watchers:
+                            # callbacks may remove the watcher, hence the copy
+                            for watcher in watchers.copy():
+                                watcher.on_connected_changed(new_connected)
 
                 elif interface == defs.GATT_CHARACTERISTIC_INTERFACE:
                     # handle characteristic value change watchers
-
                     if "Value" in changed:
-                        for device_path, _, on_value_changed in self._device_watchers:
-                            if message.path.startswith(device_path):
-                                on_value_changed(message.path, self_interface["Value"])
+                        new_value = self_interface["Value"]
+                        device_path = device_path_from_characteristic_path(message_path)
+                        watchers = self._device_watchers.get(device_path)
+                        if watchers:
+                            for watcher in watchers:
+                                watcher.on_characteristic_value_changed(
+                                    message_path, new_value
+                                )
 
-    def _run_advertisement_callbacks(
-        self, device_path: str, device: Device1, changed: Iterable[str]
-    ) -> None:
+    def _run_advertisement_callbacks(self, device_path: str, device: Device1) -> None:
         """
         Runs any registered advertisement callbacks.
 
         Args:
             device_path: The D-Bus object path of the remote device.
             device: The current D-Bus properties of the device.
-            changed: A list of properties that have changed since the last call.
         """
-        for (callback, adapter_path) in self._advertisement_callbacks:
-            # filter messages from other adapters
-            if adapter_path != device["Adapter"]:
-                continue
-
+        adapter_path = device["Adapter"]
+        for callback in self._advertisement_callbacks[adapter_path]:
             callback(device_path, device.copy())
 
 
